import threading
import datetime
from enum import IntEnum
import numpy as np
from .vsapy import *
from .vsatype import VsaBase, VsaType
from vsapy.logger_utils import *
log = setuplogs(level='INFO')

from vsapy.bag import *
from vsapy.helpers import *


class TimeStamp(object):
    @staticmethod
    def get_creation_data_time_stamp():
        return 'Timestamp: {:%Y-%m-%d %H:%M:%S.%f}'.format(datetime.datetime.now())

    @staticmethod
    def compare_time_stamps(t1, t2):
        t1obj = datetime.datetime.strptime(t1, 'Timestamp: %Y-%m-%d %H:%M:%S.%f')
        t2obj = datetime.datetime.strptime(t2, 'Timestamp: %Y-%m-%d %H:%M:%S.%f')
        return t1obj.time() == t2obj.time()


def createSymbolVectors(symbols, *args, creation_data_time_stamp=None, **kwargs):
    if creation_data_time_stamp is None:
        creation_data_time_stamp = TimeStamp.get_creation_data_time_stamp()
    # A dictionary is slightly faster than Graham's list look up but only marginally....
    sym = {"creation_data_time_stamp": creation_data_time_stamp}

    for a in symbols:
        sym[a] = vsa.randvec(*args, **kwargs)

    return sym


def create_base_vecs(start, end, veclen, ascii_names=True,
                     creation_data_time_stamp=None, vsa_type=VsaType.BSC, **kwargs):

    if creation_data_time_stamp is None:
        time_stamp = TimeStamp.get_creation_data_time_stamp()
    else:
        time_stamp = creation_data_time_stamp

    base_vecs = {"creation_data_time_stamp": time_stamp}  # Dictionary(Of String, BitArray)
    if ascii_names:
        start = ord(start)
        end = ord(end)
    for x in range(start, end + 1):  # inclusive of end
        if ascii_names:
            c = chr(x)
        else:
            c = x
        base_vecs.update({c: vsa.randvec(veclen, vsa_type=vsa_type, **kwargs)})
        list(base_vecs.items())

    return base_vecs


class RoleVecs(object):
    next_chunk_id = 0  # Used for debug to identify a particular chunk.
    symbols = " abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890.;:,_'!?-[]&*"

    def __init__(self, veclen, random_seed=None, creation_data_time_stamp=None, vsa_type=VsaType.BSC, **kwargs):
        """

        :param veclen: Dimensionality of the vectors to be created.
        :param creation_data_time_stamp:
        """
        if creation_data_time_stamp is None:
            self.creation_data_time_stamp = TimeStamp.get_creation_data_time_stamp()
        else:
            self.creation_data_time_stamp = creation_data_time_stamp

        # self.id_stamp = "NOT SET"
        # if "NOT SET" in self.id_stamp:
        #     self.id_stamp = self.creation_data_time_stamp

        np.random.seed(random_seed)
        self.symbol_dict = createSymbolVectors(RoleVecs.symbols, veclen,
                                               creation_data_time_stamp=creation_data_time_stamp,
                                               vsa_type=vsa_type, **kwargs)
        self.num_dict = create_base_vecs("0", "9", veclen, True,
                                         creation_data_time_stamp=creation_data_time_stamp, vsa_type=vsa_type, **kwargs)

        self.match_message = vsa.randvec(veclen, vsa_type=vsa_type, **kwargs)  # The random alphanumeric match-tag used in workflow requests

        self.id = vsa.randvec(veclen, vsa_type=vsa_type, **kwargs)  # The responder's vector id used in match replies to differentiate between
                                            # responders in a workflow request
                                            # (this is as an alternative to self.role_match_message)

        self.jobid = vsa.randvec(veclen, vsa_type=vsa_type, **kwargs)  # The senders job-id in a workflow request
        self.matchval = vsa.randvec(veclen, vsa_type=vsa_type, **kwargs)  # The hsim match quality in a reply msg
        self.vec_count = vsa.randvec(veclen, vsa_type=vsa_type, **kwargs)  # The number of vecs embedded in this vector
        self.stopvec = vsa.randvec(veclen, vsa_type=vsa_type, **kwargs)  # The chunk stop vector
        self.permVecs = tuple([vsa.randvec(veclen, vsa_type=vsa_type, **kwargs) for _ in range(150)])
        self.parent = vsa.randvec(veclen, vsa_type=vsa_type, **kwargs)  # Used in DAG encoding
        self.child = vsa.randvec(veclen, vsa_type=vsa_type, **kwargs)  # Used in DAG encoding
        # -------------------------------------------------------------------------
        self.tvec_tag = vsa.randvec(veclen, vsa_type=vsa_type, **kwargs)  # Tvec POSITION-Role-vector
        self.current_pindex = vsa.randvec(veclen, vsa_type=vsa_type, **kwargs)
        self.pindex_numeric_base = vsa.randvec(veclen, vsa_type=vsa_type, **kwargs)  # Used when representing numbers as a cyclic-shifted vec


# From initialisation we want certain values to agree across all instances of these vector modules
# They are initialised together here in a fixed order.
# Changing the order will stop remote services participating in this scheme from working because
# the vector alphabet and perm vectors / role vectors in class NewChunkPvecs will be different


def checkt_base_objects_are_synced(data_files, vsa_type=VsaType.BSC):
    must_init_flag = False
    data_objects = {}
    first_file_time_stamp = None
    for obj_name, fname in data_files.items():
        obj = deserialise_object(fname)
        if obj is None:
            return True, None

        for k, v in obj.__dict__.items():
            if isinstance(v, VsaBase) and v.vsa_type != vsa_type:
                return True, None

        data_objects[obj_name] = obj
        try:
            if isinstance(obj, dict):
                this_time_stamp = obj["creation_data_time_stamp"]
            else:
                # obj is a class
                this_time_stamp = obj.creation_data_time_stamp
        except (AttributeError, KeyError) as e:
            log.info(f"Creation date error on object {k}")
            must_init_flag = True
            break

        if first_file_time_stamp is None:
            first_file_time_stamp = this_time_stamp
            continue

        if not TimeStamp.compare_time_stamps(first_file_time_stamp, this_time_stamp):
            log.info(f"Object {k}: time-stamp does not match other objects time-stamps")
            must_init_flag = True
            break

    if not must_init_flag:
        log.info(f"All files appear to be in sync with time-stamp {first_file_time_stamp}")
    else:
        log.info("BASE FILES out of sync, REBUILDING")

    return must_init_flag, data_objects


def create_role_data(*args, data_files=None, vec_len=10000,
                     rand_seed=None, vsa_type=VsaType.BSC, force_new_vecs=False, **kwargs):

    """
    Loads from disk or creates new role vector containers/objects to ensure that all file objects loaded
    were created at the same time.

    If new objects are created serialises each back to disk.

    :param data_files:
    :param vec_len: Dimensionality of vectors to be created
    :param rand_seed: to always create a known set of role vectors based on the order of creation.
    :return: dictionary of created objects
    """

    if data_files is None:
        # role vector files should be added to this list
        data_files = {
            "role_vectors": "role_vectors.bin",
        }

    must_init_flag = True
    if not force_new_vecs:
        must_init_flag, data_objects = checkt_base_objects_are_synced(data_files, vsa_type)
    creation_data_time_stamp = TimeStamp.get_creation_data_time_stamp()

    if must_init_flag:
        # Newly create/recreate each object that contains role vectors that must be in sync with each other.
        role_vecs = RoleVecs(vec_len, random_seed=rand_seed,
                             creation_data_time_stamp=creation_data_time_stamp,
                             vsa_type=vsa_type, **kwargs)
        serialise_object(role_vecs, data_files['role_vectors'])
    else:
        role_vecs = data_objects['role_vectors']

    print(f"Loaded role_vectors.bin, time-stamp {role_vecs.creation_data_time_stamp}")

    return role_vecs



import math
from scipy import stats
import vsapy
from .vsatype import *


class Laiho(VsaBase):
    vsatype = VsaType.Laiho

    @classmethod
    def default_numpy_type(cls):
        """
        :return: the default numpy datatype for this type of VSA.
        """
        return 'int16'

    @classmethod
    def is_laiho_type(cls, vsa_type):
        if isinstance(vsa_type, vsapy.BagVec):
            return isinstance(vsa_type.myvec, Laiho)
        elif isinstance(vsa_type, Laiho):
            return True
        if vsa_type == VsaType.Laiho or vsa_type == VsaType.LaihoX:
            return True
        return False

    @classmethod
    def expandbits(cls, v):
        # Init sparse bound vector (s_bound) with zeros
        s_bound = np.zeros((v.slots, v.bits_per_slot))  # Create a slotted vector with

        slot_index = -1
        for bit_index in v:
            slot_index += 1
            s_bound[slot_index][bit_index] = 1

        return s_bound

    @classmethod
    def repackbits(cls, v):
        return VsaBase(np.argmax(v, axis=1), vsa_type=VsaType.Laiho, bits_per_slot=len(v[0]))

    @classmethod
    def unpackbits(cls, v):
        # Laiho vecs are manipulated in packed format.
        return v

    @classmethod
    def packbits(cls, v):
        # Laiho vecs are manipulated in packed format.
        return v

    @classmethod
    def random_threshold(cls, *args, p=0, stdev_count=4.4, **kwargs):
        """
        Should return a normalised value of the similarity match that would be expected when comparing
        random/orthorgonal vectors.

        :param args: A sample vector the dimensions of which should be used to calculate threshold.
        :param stdev_count:
        :param kwargs: slots=<int>, bits_per_slot=<int>
        :return: normalised threshold value
        :rtype: float
        """
        if len(args) > 0:
            if isinstance(args[0], Laiho):
                slots = len(args[0])
                bits_per_slot = args[0].bits_per_slot
            else:
                raise ValueError("Expected a Laiho/LaihoX vector.")
        else:
            slots = kwargs.get('slots', -1)
            if slots < 1:
                raise ValueError("You must supply a sample vector, or set optional parameter 'slots'")
            bits_per_slot = kwargs.get('bits_per_slot', -1)
            if bits_per_slot < 1:
                raise ValueError("You must supply a sample vector, or set optional parameter 'bits_per_slot'")

        if p <= 0:
            p = 1 / bits_per_slot  # Probability of a match between random vectors

        return p + stdev_count * math.sqrt(p*(1-p)/slots)

    @classmethod
    def randvec(cls, dims, word_size=16,  vsa_type=VsaType.Laiho, bits_per_slot=None):
        """
        :param bits_per_slot:
        :param dims: integer or tuple, specifies shape of required array, last element is no bits per vector.
        :param word_size: numpy's word size parameter, e.g. for BSCs wordsize=8 becomes 'uint8'.
        :param vsa_type: type of VSA subclass to create from VsaType class.
        :return: a matrix of vectors of shape 'dims'.
        """
        return VsaBase(np.random.randint(0, bits_per_slot, size=dims, dtype='int' + str(word_size)),
                       vsa_type=vsa_type,
                       bits_per_slot=bits_per_slot)

    @classmethod
    def bind(cls, a, b):  # actually bind/unbind for binary and ternary vecs
        """
        Comutative binding operator
        :param a: VSA vec
        :param b: VSA vec
        :return: vector associating/coupling a to b that is dissimilar to both a and b.
                 In most cases bind(a, b) is analogues to multiplication, e.g. bind(3,4)=>12.
                 If we know one of the operands we can recover the other using unbind(a,b) e.g unbind(3,12)=>4
        """
        assert a.vsa_type == b.vsa_type, "Mismatch vsa_types"
        assert a.bits_per_slot == b.bits_per_slot, "Bits per slot mismatch"
        assert a.slots == b.slots, "not of slots mismatch"
        return (a + b) % a.bits_per_slot

    @classmethod
    def unbind(cls, a, b):
        """
        Non-Comutative unbinding operator. Decouples a from b and vice-versa. The result
        :param a: is the bundled vector
        :param b: is the role or filler vector
        :return: reverses a bind operation. If z = bind(x, y) then x = unbind(z, y) and y = unbind(z, x).
                 The return is orthogonal to x nd y if x and y have not been previously associated with bind(x, y).
        """

        # raise NotImplementedError('Laiho unbunding is NonCommutative, use self.unbind()')
        # return
        assert a.vsa_type == b.vsa_type, "Mismatch vsa_types."
        assert a.bits_per_slot == b.bits_per_slot, "Bits per slot mismatch."
        assert a.slots == b.slots, "Number of slots mismatch."

        return (a - b) % a.bits_per_slot

    def cyclic_shift(self, shift):
        Zs = []
        for i in range(0, self.slots):
            Zs.append(self[(self.slots - shift + i) % self.slots])

        Zs1 = [self[(self.slots - shift + i) % self.slots] for i in range(self.slots)]
        Zs2 = np.roll(self, shift)
        return Zs

    @classmethod
    def sum2(cls, vlist, *args, **kwargs):
        """
        Maintains vsa_type custom attribute when perfoming numpy.sum()
        Todo: there is probably a better way than this.
        """
        # Super slow method compared to sum() but was used as a sanity checker.
        unpacked_list = [v.expandbits(v).flatten() for v in vlist]
        orvec = np.sum(unpacked_list, axis=0).reshape(vlist[0].slots, vlist[0].bits_per_slot)
        outvec = np.ma.masked_array(orvec, (orvec == 0)).argmax(axis=1)

        return VsaBase(np.array(outvec), vsa_type=VsaType.Laiho, bits_per_slot=vlist[0].bits_per_slot)

    @classmethod
    def sum1(cls, vlist, *args, **kwargs):
        """
        Maintains vsa_type custom attribute when perfoming numpy.sum()
        Todo: there is probably a better way than this.
        """

        # This method is 2X slower than helpers.py mode() !
        return VsaBase(stats.mode(vlist)[0][0], vsa_type=VsaType.Laiho, bits_per_slot=vlist[0].bits_per_slot)

    @classmethod
    def sum(cls, vlist, *args, **kwargs):
        """
        Maintains vsa_type custom attribute when perfoming numpy.sum()
        Todo: there is probably a better way than this.
        """
        # helpers.py mode() is 2x faster than scipy.stats.mode() !
        return VsaBase(mode(vlist, axis=0)[0], vsa_type=VsaType.Laiho, bits_per_slot=vlist[0].bits_per_slot)

    @classmethod
    def normalize(cls, sv, seqlength=None, rv=None):
        """
        Normalize the VSA vector
        :param sv: input VSA vector
        :param seqlength: Optional, the number of vectors created by summing a sequence of vectors.
                          This is required for BSCs, if not set, seqlength=None will raise and error .
        :param rv: Optional random vector, used for splitting ties on binary and ternary VSA vectors.
        :return: new VSA vector
        """

        return VsaBase(np.array([np.argmax(sv[s]) for s in range(0, sv.shape[0])]),
                       vsa_type=VsaType.Laiho, bits_per_slot=sv.shape[1])

    @classmethod
    def hdist(cls, a, b):
        """
        :param a: vsa vector
        :param b: vsa vector
        :return: normalized hamming distance between a and b. 0.0=exact match.
        """
        return 1.0 - cls.hsim(a, b)

    @classmethod
    def hsim(cls, a, b):
        """
        :param a: vsa vector
        :param b: vsa vector
        :return: normalized hamming distance between a and b. 1.0=exact match.
        """
        assert a.vsa_type == b.vsa_type, "Mismatch vsa_types"
        if isinstance(b, Laiho):
            match = np.count_nonzero(a == b)
            return match / a.slots

        raise ValueError("Mismatch vsa_types")

    @staticmethod
    def slots_from_bsc_vec(bsc_dim, bits_per_slot):
        return int(bsc_dim / math.log(bits_per_slot, 2))

    @staticmethod
    def bits_from_bsc_vec(bsc_dim, slots):
        return int(2 ** (bsc_dim/slots))

    @staticmethod
    def bsc_dim_from_laiho(slots, bits_per_slot):
        return int(slots * math.log(bits_per_slot, 2))


def mode(ndarray, axis=0):
    # Check inputs
    ndarray = np.asarray(ndarray)
    ndim = ndarray.ndim
    if ndarray.size == 1:
        return (ndarray[0], 1)
    elif ndarray.size == 0:
        raise Exception('Cannot compute mode on empty array')
    try:
        axis = range(ndarray.ndim)[axis]
    except:
        raise Exception('Axis "{}" incompatible with the {}-dimension array'.format(axis, ndim))

    # If array is 1-D and numpy version is > 1.9 np.unique will suffice
    if all([ndim == 1,
            int(np.__version__.split('.')[0]) >= 1,
            int(np.__version__.split('.')[1]) >= 9]):
        modals, counts = np.unique(ndarray, return_counts=True)
        index = np.argmax(counts)
        return modals[index], counts[index]

    # Sort array
    sort = np.sort(ndarray, axis=axis)
    # Create array to transpose along the axis and get padding shape
    transpose = np.roll(np.arange(ndim)[::-1], axis)
    shape = list(sort.shape)
    shape[axis] = 1
    # Create a boolean array along strides of unique values
    strides = np.concatenate([np.zeros(shape=shape, dtype='bool'),
                                 np.diff(sort, axis=axis) == 0,
                                 np.zeros(shape=shape, dtype='bool')],
                                axis=axis).transpose(transpose).ravel()
    # Count the stride lengths
    counts = np.cumsum(strides)
    counts[~strides] = np.concatenate([[0], np.diff(counts[~strides])])
    counts[strides] = 0
    # Get shape of padded counts and slice to return to the original shape
    shape = np.array(sort.shape)
    shape[axis] += 1
    shape = shape[transpose]
    slices = [slice(None)] * ndim
    slices[axis] = slice(1, None)
    # Reshape and compute final counts
    counts = counts.reshape(shape).transpose(transpose)[tuple(slices)] + 1

    # Find maximum counts and return modals/counts
    slices = [slice(None, i) for i in sort.shape]
    del slices[axis]
    index = np.ogrid[slices]
    index.insert(axis, np.argmax(counts, axis=axis))
    return sort[tuple(index)], counts[tuple(index)]


import re
from vsapy.bag import *
from vsapy.cspvec import *
from vsapy.helpers import *


class VsaTokenizer(object):
    def __init__(self, role_vecs, _usechunksforwords=True,
                 allow_skip_words=False, skip_words=None, skip_word_criterion=None,
                 use_word2vec=False, word2Vec_model=None, my_r2b=None):

        """
        This is a helper class in containing a number of parsers that can encode words, sentences and whole
        documents into a vsa vector or set of hierarchical vectors.

        :param symbol_dict:
        :param _usechunksforwords:
        :param allow_skip_words:
        :param skip_words:
        :param skip_word_criterion: function returning bool indicates whether to skip this word, can pass as a lambda
        :param use_word2vec:
        :param word2Vec_model: A word2vec semantic vector model.
        :param my_r2b:
        """
        self.role_vecs = role_vecs
        self.symbol_dict = role_vecs.symbol_dict
        self.use_word2vec = use_word2vec
        self.miss_from_word2vec = {}
        self.linecheck = []  # Used for debug to check vsa unbinding replays. Allows us to check if the unbinding
                             # reproduces the same document lines as were built duing encoding.
        self.total_word_count = 0  # Used to count the total number of words seen by the tokenizer
                                   # Reset this before starting a new document.

        if use_word2vec:
            assert word2Vec_model is not None, "you must supply a word2vec model"
            assert my_r2b is not None, "you must supply a RealToBinary converter"
            self.model = word2Vec_model
            self.my_r2b = my_r2b
        else:
            self.model = None  # word to vec model
            self.my_r2b = None  # real to binary converter

        self.allow_skips = allow_skip_words
        if allow_skip_words:
            assert skip_words is not None, "you must specify a skip words dictionary"
            self.skip_words = skip_words
        else:
            self.skip_words = {}

        self.meets_skip_words_criterion = skip_word_criterion

        self.seen_words = {}  # Tracks seen words so they do not need to be rebuilt

        self.usechunksforwords = _usechunksforwords
        if self.usechunksforwords:
            self.createWordVector = self.chunkWordVector  # Using our own chunking scheme, allows individual words as services
        else:
            self.createWordVector = self.gb_wordvector_as_chunk  # Graham's piBinding - faster

    @staticmethod
    def replace_non_alphanumerics(source, replacement_character='_'):
        result = re.sub("[^_' a-zA-Z0-9]", replacement_character, source)

        return result

    def chain_vecs(self, word):
        chain = np.roll(self.symbol_dict[word[0]], 1)
        for shift, c in enumerate(word[1:], 2):
            chain = vsa.bind(chain, np.roll(self.symbol_dict[c], shift))

        return chain

    def chunkWordVector(self, word):
        lettervecs = [self.symbol_dict[a] for a in word]

        cnk = CSPvec(word, lettervecs, self.role_vecs)
        if self.usechunksforwords:
            self.linecheck.append(word)  # record the word for verification

        return cnk

    def createWordVector_GB(self, word):
        shift = 0
        letter_vecs = []
        for c in word:
            shift += 1
            letter_vecs.append(np.roll(self.symbol_dict[c], shift))

        return vsapy.BagVec(letter_vecs).myvec

    def gb_wordvector_as_chunk(self, word):
        return CSPvec(word, [self.createWordVector_GB(word)], self.role_vecs)

    def get_word2Vec(self, w):
        '''
        Returns None if word not found in Model, otherwise BSC mapped from model's real number vector
        '''
        w1 = w
        v = None
        if w not in self.model:
            if w1[0] == "'":
                w1 = w[1:]  # Chop off a leading quote '
            if len(w1) > 0 and w1[-1] == "'":
                w1 = w[:-1]  # Chop off a trailing quote '
            if len(w1) > 2 and w[-2] == "'":
                w1 = w.replace("'d", "ed")
                w1 = w1.replace("'s", "")

        if w1 in self.model:
            v = self.my_r2b.to_bin(self.model[w1])

        return v

    def get_word_vector(self, w):
        v = None
        try:
            if self.use_word2vec:
                v = self.get_word2Vec(w)

            if v is None:
                v = self.createWordVector(w).myvec
                if self.use_word2vec:
                    # Count the number of words occurances not in the word2vec model
                    self.miss_from_word2vec[w] = self.miss_from_word2vec.get(w, 0) + 1

        except Exception as e:
            e = e

        return v

    def get_vector_or_skip_the_word(self, w):
        '''
        Returns BSC vector representation of the word or None if failed to build a vector
        (e.g word is in skip list or word contains all non-printing chars.
        '''
        v = None
        try:
            if self.allow_skips and (w in self.skip_words or self.meets_skip_words_criterion(w)):
                if w not in self.skip_words:
                    self.skip_words[w] = PackedVec(self.get_word_vector(w))
                return None

            if w in self.seen_words:
                if w in self.miss_from_word2vec:
                    self.miss_from_word2vec[w] += 1  # Count the occurrences of words missing from our word2vec model
                return self.seen_words[w].myvec
            else:
                v = self.get_word_vector(w)
                if v is not None:
                    self.seen_words[w] = PackedVec(v)  # We only build new word vectors once.

        except [ValueError, IndexError] as e:
            e = e

        return v

    def chunkSentenceVector(self, sentence1):

        db_sentence = None
        wordvecs = []
        try:
            sentence = sentence1.replace('--', ' ')
            sentence = VsaTokenizer.replace_non_alphanumerics(sentence, ' ')
            words = sentence.split()

            for w in words:
                w = w.strip()
                w = w.replace("'d", "ed")
                if w is None: continue

                vec = self.get_vector_or_skip_the_word(w)
                if vec is not None:
                    self.total_word_count += 1
                    wordvecs.append(CSPvec(w, [vec], self.role_vecs))

            if len(wordvecs) == 0:
                wordvecs = wordvecs  # For Debug
                return None
            else:
                try:
                    db_sentence = CSPvec.buildchunks(sentence, wordvecs, self.role_vecs,
                                                     split_tail_evenly=True, rebuild_names=True)
                except [ValueError, IndexError] as e:
                    print(e)
                    db_sentence = db_sentence  # For Debug

        except [ValueError, IndexError] as e:
            print(e)
            e = e  # For Debug

        return db_sentence




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vsapy']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.1.1,<10.0.0',
 'colorama>=0.4.5,<0.5.0',
 'gmpy2>=2.1.2,<3.0.0',
 'matplotlib>=3.5.2,<4.0.0',
 'nltk>=3.7,<4.0',
 'numpy>=1.22.4,<2.0.0',
 'pandas>=1.4.2,<2.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'regex>=2022.4.24,<2023.0.0',
 'scikit-image>=0.19.3,<0.20.0',
 'scipy>=1.8.1,<2.0.0',
 'setuptools>=62.3.2,<63.0.0',
 'six>=1.16.0,<2.0.0',
 'unicodedata2>=14.0.0,<15.0.0',
 'wheel>=0.37.1,<0.38.0',
 'xmltodict>=0.13.0,<0.14.0']

setup_kwargs = {
    'name': 'vsapy',
    'version': '0.7.8',
    'description': 'Vector Symbolic Architecture(VSA) library that allows building VSA apps that use various flavours of VSA vectors.',
    'long_description': '# vsapy - Vector Symbolic Architecture(VSA) library.\nThis library implements the common methods used in hyperdimensional computing/Vector Symbolic Architectures. Namely\n`bind`, `unbind` and some bundling operations, `bag`, `ShiftedBag`, `NamedBag` and a hierarchical bundling \nmethod `CSPvec`. The main objective of the library is to enable a single API to cater for various flavours of VSA, \nincluding `Binary Spatter Codes`, `Ternary`, `TernaryZero`, `Laiho` and `LaihoX` (a simplified `Laiho` that \nis faster and supports incremental bundling without catastrophic forgetting). \n\nA set of demo test cases are supplied to demonstrate the use of the library calls.\n\n\n## If installing from PyPi simply\n  - Poetry add vsapy\n<br/>or<br/>\n  - pip install vsapy (into your environment)\n\n## Installing from source\n  - clone the code to a directory of your choice, say "vsapy"\n\n### Installing Dependancies  \n- Poetry: the easiest way is using poetry\n  - cd vsapy\n  - poetry install\n  - poetry shell  (to activate the environment)\n  \n\n- pip install vsapy\n  - create an environment using your favorite environment manager\n  - e.g. conda create -n vsapy39 python=3.9\n  - conda activate vsapy39\n  - pip install -r requirements.txt\n\n### Usage\nHint: Valid values for `VSaType` are, `VsaType.BSC`, `VsaType.Laiho`, `VsaType.LaihoX`(fastest), `VsaType.Tern`, \n`VsaType.TernZero` and `VsaType.HRR`\\\n(** Note, the demos listed below will not run with type `VsaType.HRR` **). <br/><br/>\n\n\n- For examples of using the vsapy library, see the code examples in the ./tests directory. Note there are no \ncommand-line arguments implemented for the tests at the moment. To change the type of VSA in use, edit the code changing\n`vsa_type=VsaType.BSC` as mentioned below. All of the test cases can be run by simply invoking from the command line, \ne.g., `$python cspvec_sequence.py`.\n\n\n\n  - `cspvec_sequence.py`: This is the most straightforward demo. Try this first. It demonstrates building a sentence as \na vsa sequence and stepping forward & backwards. Change `vsa_type=VsaType.BSC` in the code to change the type of VSA\nused to build the representation. <br/><br/>\n  \n  - `build_docs.py`: demonstrates combining large documents into a hierarchical vsa code book. The top-level document \nvector is a high-level semantic representation of the entire document. Change `vsa_type=VsaType.BSC` in the code to \nchange the type of VSA used to build the representation. <br/><br/>\n\n    - `load_docs.py`: compares the document vectors built using `build_docs.py` at various levels in the \ndocument hierarchy. <br/><br/> Change `levels_to_extract = [0, 1]`, `0=top-level document vectors`, `1=Act-level vectors`, \n`2=Scene-level vectors` and so on (Can set to any level, e.g., `levels_to_extract = [2]` would compare only \nScene-level vectors). <br/><br/>\n\n      - Understanding output names: `OE_=Old English`, `NE_=New English`, `nf_=NoFear Shakespeare`, `tk_=NLTK Shakespeare`,\n`og_=Alternate Shakespeare`, `ham=Shakespeare\'s Hamlet` , `mbeth=Shakespeare\'s Macbeth`. <br/><br/> \n\n  - `json2vsa.py`: demonstrates the creation of a VSA vector from an input JSON file and shows a comparison of various\nJSONs using VSA. Change `vsa_type=VsaType.BSC` in the code to change the type of VSA used to build the representation.\n\n\n',
    'author': 'Chris Simpkin',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/vsapy/vsapy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import random

import vsapy
from vsapy.role_vectors import *
from vsapy.bag import *
from vsapy.logger_utils import *
log = setuplogs(level='INFO')


class CSPvec(RawVec):
    id_stamp = "NOT SET"
    mythreadlock = threading.RLock()
    next_chunk_id = 0
    break_on_chunk_id = -1
    trace_threshold = 0.013  # 0.53  # 0.547 #0.528  #0.53  # 0.525  # 0.54  # 0.532

    # ------------------------------------------------------------------------
    # If we change this list we MUST update calc_match_level():
    # trace_thresholds = [0.8, 0.7, 0.6, 0.54, 0.53, 0.525]
    # trace_thresholds = [0.6, 0.59, 0.58, 0.57, 0.56, 0.55, 0.54, 0.53, 0.525]
    # trace_thresholds = [0.6, 0.56, 0.54]
    trace_thresholds = [0.53]
    # trace_thresholds = [0.8, 0.7, 0.54]

    def __init__(self, name, veclist, role_vecs, chunks=None, creation_data_time_stamp=None):

        self.creation_data_time_stamp = role_vecs.creation_data_time_stamp

        self.aname = name
        if isinstance(veclist[0], CSPvec):
            # This means we are building from a list of previously created chunks
            self.__terminal_node = False
            self.chunklist = veclist  # the real vectors, for debug and possible use in clean up
            veclist = [c.myvec for c in veclist]
        else:
            self.__terminal_node = True
            # veclist = veclist

        self.chunksize = len(veclist)
        self.veclen = len(veclist[0])

        self.roles = role_vecs
        self.permVecs = role_vecs.permVecs

        self.chunk_id = CSPvec.next_chunk_id
        with CSPvec.mythreadlock:
            CSPvec.next_chunk_id += 1

        self.chunklist = chunks  # the real vectors, for debug and possible use in clean up
        self.start_tag_vec = None

        # The stopvec is made from the last vector in the list this ensures that we get better matches
        # the main reason for this is to ensure that the last child vector in the list is able
        # to detect that the next vector is a stop vector.
        # The spinoff benefit is that we get better matches because vector lists ending in the same
        # vector will not differ by the stop vector.
        self.stopvec = vsa.bind(np.roll(veclist[-1], 1), self.roles.stopvec)

        if self.isTerminalNode:
            # we are creating a basic compound vector these do not need a stop vector, however if the number of vectors
            # to be added is even we add one enyway to make the majority vote work niceley
            if self.chunksize % 2 == 0:
                # For terminal nodes this stop_vec is actually a random vec but lets us keep an exact count
                # of the vectors in the compound vec
                veclist.append(self.stopvec)
        else:
            veclist.append(self.stopvec)

        permed_vec_list = self.permveclist(veclist)
        super(CSPvec, self).__init__(permed_vec_list)

        if self.isTerminalNode:
            # If we are a terminal node we will never look for our own stop_vec. However, we do want to detect the
            # parent vector's stop_vec.  We therefore pre-set the stop_vec to match with self.myvec because, if this
            # vec is in the last position of a higher-level list, self.myvec will be used to build the parent's stopvec.
            self.stopvec = vsa.bind(np.roll(self.myvec, 1), self.roles.stopvec)
    

    def permveclist(self, veclist):
        """
        Adding like this, p0 * a^1 + (p0 * p1 * b^2) + (p0 * p1 * p2 * c^3) + ... Where '^' means cyclic-shift
        Using this method we gain benefit because we will get better similarity matchups
        since sequence is controlled by a fixed set of random vectors

        :param veclist:
        :return: Un-normalised_vec, sub-vec_cnt, normalised-vec
        """

        if len(veclist) == 1:
            return veclist

        try:
            pindex = 0
            piv = self.permVecs[0]
            #sumvec = vsa.bind(piv, np.roll(veclist[0], pindex + 1))
            role_filler_list = [vsa.bind(piv, np.roll(veclist[0], pindex + 1))]
            cnt = 1
            for y in veclist[1:]:
                cnt += 1
                pindex += 1
                piv = vsa.bind(piv, self.permVecs[pindex])
                v = vsa.bind(piv, np.roll(y, pindex + 1))
                role_filler_list.append(v)

        except IndexError as e:
            pindex = pindex  # For Debug
            raise IndexError(e)

        return role_filler_list

    @property
    def isTerminalNode(self):
        return self.__terminal_node

    @classmethod
    def createchunk(cls, name, chunks, rolevecs):
        return cls(name, chunks, rolevecs, chunks)

    @classmethod
    def createchunk_build_name_from_chunks(cls, name, chunks, role_vecs):
        name = ' '.join([c.aname for c in chunks])
        return cls(name, chunks, role_vecs, chunks)

    @classmethod
    def bag_chunks(cls, name, chunks, rolevecs):
        return NamedBag(name, chunks)

    @classmethod
    def buildchunks(cls, name, chunklist, role_vecs, split_tail_evenly=True, maxvecs=90, rebuild_names=False, new_chunker=None):
        """

        :param name: Name of this chunk
        :param chunklist: list of individual vectors
        :param split_tail_evenly: True=tails of list split approx in half, else straglers just moppoed up into one vec
        :return: top level chunk of a heirachy built from the flat list of chunks supplied
        """

        assert chunklist is not None, "chunklist is not empty"
        if isinstance(chunklist[0], CSPvec):
            level = 1
        else:
            level = 0

        if not new_chunker:
            if rebuild_names:
                create_chunk = cls.createchunk_build_name_from_chunks
            else:
                create_chunk = cls.createchunk
        else:
            create_chunk = new_chunker

        # m = 0
        llen = len(chunklist)
        if llen > maxvecs:
            while llen > maxvecs:
                chnks = []
                part_id = 0
                m = 0
                while m + maxvecs <= llen:
                    # when the remaining list is less than two chunks we want to split it equally
                    if (llen - m) % maxvecs != 0 and split_tail_evenly and (llen - m < maxvecs * 2):
                        break
                    sub_chunk_list = chunklist[m:m + maxvecs]
                    chnks.append(create_chunk("{}, L{:02d}, part{:02d}".format(name, level, part_id),
                                              sub_chunk_list, role_vecs))
                    m += maxvecs
                    part_id += 1

                if m < llen:  # if there is a tail left to chunk
                    if split_tail_evenly:
                        # we want to split the remainder into two fairly even chunks
                        msplit = int((llen - m) / 2)
                        chnks.append(create_chunk(f"{name}, L{level:02d}, part{part_id:02d}",
                                                  chunklist[m:m + msplit], role_vecs))
                        m += msplit
                        part_id += 1
                        chnks.append(create_chunk(f"{name}, L{level:02d}, part{part_id:02d}",
                                                  chunklist[m:llen], role_vecs))
                    else:
                        # grab any left over vecs
                        chnks.append(create_chunk(f"{name}, L{level:02d}, part{part_id:02d}",
                                                  chunklist[m:llen], role_vecs,))

                level += 1  # sub chunks were built so next level is at one level higher than its children
                if len(chnks) <= maxvecs:
                    chunklist = chnks
                    break
                else:
                    llen = len(chnks)
                    chunklist = chnks[:]

        dbcnks = create_chunk(name, chunklist, role_vecs)
        return dbcnks

    def unbind_one_step(self, invec, pindex):
        return np.roll(vsa.unbind(invec, np.roll(self.roles.permVecs[pindex], -1 * pindex)), -1)

    def recover_stop_vec_from_vec_count(self, invec, vec_count):
        invec1 = invec[:]
        for pindex in range(vec_count):
            invec1 = self.unbind_one_step(invec1, pindex)

        return invec1

    def as_seq(self, im_the_requester=False):

        self.im_the_requester = im_the_requester

        # We write the total number of sub-vectors into the vector, if the number of sub-vecs in self.vec_cnt
        # is even, then the total number of vecs will be extended by 1 during normalisation.

        vec_count = self.vec_cnt + 2
        if not isinstance(self.myvec, vsa.Laiho):
            if vec_count % 2 == 0:
                vec_count += 1  # An extra vec is added during normalization.
        sub_vecs = [
            self.rawvec,
            self.roles.tvec_tag,  # metadata, identifies the start of a sequence vector
            np.roll(self.roles.vec_count, vec_count),  # metadata, number of sub-vecs inn this vec including metadata
        ]

        bagvec = BagVec(sub_vecs, vec_count)
        if isinstance(self.myvec, vsa.Laiho):
            seq_vec = VsaBase(bagvec.myvec, vsa_type=bagvec.myvec.vsa_type, bits_per_slot=sub_vecs[0].bits_per_slot)
        else:
            seq_vec = VsaBase(bagvec.myvec, vsa_type=bagvec.myvec.vsa_type)

        start_vec = self.unbind_one_step(seq_vec, 0)  # Zero is the start pindex

        log.info(f"{self.chunk_id:04d}: START_VEC: NoSubVecs({vec_count}) vec_len={len(start_vec)} | {self.aname}")

        return start_vec

    def match_theshold(self, invec):
        if isinstance(invec, vsapy.Laiho):
            M = len(invec)

            exprv = 1 / invec.bits_per_slot  # Probability of a match between random vectors
            var_rv = M * (exprv * (1 - exprv))  # Varience (un-normalised)
            std_rv = math.sqrt(var_rv)  # Stdev (un-normalised)
            hdrv = M / invec.bits_per_slot + 4.4 * std_rv  # Un-normalised hsim of two randomvectors adjusted by 'n' stdevs

            # expected_hd = 1 / invec.bits_per_slot  # Normalised expeced value
            # stdev = math.sqrt(len(invec) * expected_hd * (1 - expected_hd)) / len(invec)
            return hdrv/M
        elif invec.vsa_type == VsaType.HRR:
            stdev = math.sqrt(1 / len(invec) * (101 / 100))  # Assuming max 100 vectors
            return 0 + 4.4 * stdev  # mean=0 + 4.4 * stdev
        else:
            return 0.53

    def check_for_activation(self, invec, myvec, required_threshhold):

        if myvec is None:
            return 0, invec, -1

        match = vsa.hsim(invec, myvec)
        # TODO: change check to match the current threshold requiements - this will give us some speed up.
        if match < required_threshhold:
            return match, invec, -1, invec
        else:
            return self.new_get_next_vector(invec, match)

    def get_current_permutation(self, invec):
        if self.check_for_start_tag_vec(invec):
            return -1
        #match_threshold = self.match_theshold(invec)
        #match_threshold = VsaBase.random_threshold(invec)
        match_threshold = vsa.random_threshold(invec)

        pindex = 0
        pvec = self.roles.tvec_tag.copy()
        for p in self.roles.permVecs:
            pvec = self.unbind_one_step(pvec, abs(pindex))
            hd = vsa.hsim(pvec, invec)
            if hd >= match_threshold:
                # We know if we are better than threshold that we have the best match on role_posn finder
                # because only one pvec.role_posn finder combo will ever be a match
                break
            pindex -= 1
        if abs(pindex) >= len(self.roles.permVecs):
            log.error("ERROR: {} Could not find 'T' vector in input vector".format(self.aname))

        return abs(pindex)

    def reverse_unbind(self, invec, debug_prev_vec=None):
        pindex = self.get_current_permutation(invec)
        reverse_vec = np.roll(vsa.bind(invec, np.roll(self.roles.permVecs[pindex], -1 - pindex)), 1)

        if debug_prev_vec is not None:
            rhd = vsa.hsim(debug_prev_vec, reverse_vec)
        return reverse_vec, pindex - 1

    def forward_unbind(self, invec):
        try:
            pindex = 0 - self.get_current_permutation(invec) - 1
            nextvec = np.roll(vsa.unbind(invec, np.roll(self.roles.permVecs[0 - pindex], pindex)), -1)
        except IndexError:
            invec = invec

        return nextvec, pindex

    def wind_vec_forward(self, invec, dest_pindex, debug_target_vec=None):
        invec_pindex = self.get_current_permutation(invec)
        if invec_pindex == dest_pindex:
            return invec

        unbind = self.forward_unbind
        if invec_pindex > dest_pindex:
            unbind = self.reverse_unbind

        nextvec = invec.copy()
        while True:
            nextvec, pindex = unbind(nextvec)
            if abs(pindex) == dest_pindex:
                break

        return nextvec

    def new_get_next_vector(self, invec, hd_match):

        try:
            # pvec = np.roll(NewChunkPvecs.permVecs[0 - pindex], pindex)
            nextvec, pindex = self.forward_unbind(invec)
        except IndexError:
            log.error("{:04d}: {} - Could not resolve Tvec to calc next vec".format(self.chunk_id, self.aname))
            # We return 0 instead of hd_match to indicate that the match was unrelaible
            # and should be discounted as a valid match when check_for_activation() returns
            return 0, invec, -1, invec

        try:
            # pvec = np.roll(NewChunkPvecs.permVecs[0 - ppindex], ppindex)
            nextvec2, _ = self.forward_unbind(nextvec)
        except IndexError:
            log.error("{:04d}: {} - Could not resolve Tvec to calc next vec".format(self.chunk_id, self.aname))
            # We return 0 instead of hd_match to indicate that the match was unrelaible
            # and should be discounted as a valid match when check_for_activation() returns
            return 0, invec, -1, invec

        return hd_match, nextvec, -1 - pindex, nextvec2

    @staticmethod
    def get_numeric(datavec, role_vec, stop_at):

        i = 0
        for i in range(stop_at):
            if vsa.hsim(np.roll(role_vec, i), datavec) >= CSPvec.trace_threshold:
                break

        return i

    def get_next_vector(self):
        pvec = np.roll(CSPvec[0][self.next_vec + 1], -1 * self.next_vec)
        nextvec = np.roll(vsa.bind(pvec, self.commandvec), -1)
        return nextvec

    def check_for_stopvec(self, invec):
        return vsa.hsim(self.stopvec, invec) > self.match_theshold(invec)

    def check_for_start_tag_vec(self, invec):
        return vsa.hsim(self.roles.tvec_tag, invec) > self.match_theshold(invec)

    def flattenchunkheirachy(self, allchunks, skip_worker_chunks=False):
        if self.isTerminalNode:
            if skip_worker_chunks:
                return
            else:
                allchunks.append(self)
            return
        else:
            allchunks.append(self)
            for c in self.chunklist:
                c.flattenchunkheirachy(allchunks, skip_worker_chunks)

    def get_flat_list_of_vectors_from_chunkheirachy(self, all_vecs):
        # We only want vectors that represent services
        # So we are picking each self.myvec entry from every chunk
        if self.isTerminalNode:
            all_vecs.append(self.myvec)
            return
        else:
            all_vecs.append(self.myvec)
            for c in self.chunklist:
                c.get_flat_list_of_vectors_from_chunkheirachy(all_vecs)

